
local Carousel = LibStub("AceAddon-3.0"):GetAddon("Carousel")
if not Carousel then return end

local BarProto = CreateFrame"Frame"
Carousel.BarProto = BarProto
local meta = {__index = BarProto}
local tinsert, tremove = table.insert, table.remove

function Carousel:CreateNewFrame(name, frame)
	frame = frame or setmetatable(CreateFrame("Frame", nil, UIParent), meta)
	frame.name = name
	frame.docks = {}
	frame:Show()
	frame:EnableMouse(true)
	frame:SetScript("OnDragStart", BarProto.OnDragStart)
	frame:SetScript("OnDragStop", BarProto.OnDragStop)
	return frame
end

do
	local cache = setmetatable({}, {__mode = "k"})
	function BarProto:CreateDestroyDocks()
		local has, need = #self.docks, #Carousel.db.profile.frames[self.name].objectOrder
		if need < has then
			for i = need+1, has do
				local dock = self.docks[i]
				dock:Disable()
				self.docks[i] = nil
				cache[dock] = true
			end
		elseif need > has then
			for i = has+1, need do
				local dock = next(cache)
				if dock then
					cache[dock] = nil
					dock:Enable(i, self.name)
				else
					dock = Carousel:CreateNewDock(i, self.name)
				end
				self.docks[i] = dock
			end
		end
		for i,dock in pairs(self.docks) do
			dock.index = i
		end
	end	
end	

function BarProto:Dispose()
	self:CreateDestroyDocks()
	self:Lock()
	self:Hide()
end	

function BarProto:AddDock()
	local index = #Carousel.db.profile.frames[self.name].objectOrder+1
	Carousel.db.profile.frames[self.name].objectOrder[index] = {}
	self:Update()
	for i,dock in pairs(self.docks) do
		dock:ResetActive()
	end
	return index
end

function BarProto:RemoveDock(index)
	tremove(Carousel.db.profile.frames[self.name].objectOrder, index)
	self:Update()
	local rem = {}
	for obj, tab in pairs(Carousel.objectPositions) do
		if tab.bar == self.name then
			if tab.dock == index then
				rem[obj] = true
				tab.bar = nil
				tab.dock = nil
				tab.place = nil
			elseif tab.dock > index then
				tab.dock = tab.dock - 1
			end
		end
	end
	for i,dock in pairs(self.docks) do
		dock:ResetActive()
	end
	return rem
end	

function BarProto:AnchorDocks()
	local space = Carousel.db.profile.frames[self.name].space
	local offset = Carousel.db.profile.frames[self.name].widthOffset
	local num = #self.docks
	if num > 0 then
		self.docks[1]:ClearAllPoints()
		if Carousel.db.profile.frames[self.name].locked then
			self.docks[1]:SetPoint("LEFT", self, offset/2, 0)
		else
			self.docks[1]:SetPoint("LEFT", self, offset/2 + 8, 0)
		end
		for i = 2, num do
			self.docks[i]:ClearAllPoints()
			self.docks[i]:SetPoint("LEFT", self.docks[i-1], "RIGHT", space, 0)
		end
	end
end	

function BarProto:Update()
	self:CreateDestroyDocks()
	self:UpdateLock()
	self:AnchorDocks()
	self:UpdateOrder()
	self:UpdateScale()
	self:UpdateAnchor()
	self:UpdateHeight()
	self:UpdateLayout()
	self:UpdateTextIconPosition()
	self:UpdateBackdrop()
	self:UpdateFont()
	self:UpdateTextAlignment()
	self:UpdateTextColor()
	self:UpdateWidth()
end

function BarProto:UpdateAnchor()
	local db = Carousel.db.profile.frames[self.name]
	self:ClearAllPoints()
	self:SetPoint(db.anchorSide, nil, db.frameX/db.scale, db.frameY/db.scale)
end

function BarProto:UpdateScale()
	local scale = Carousel.db.profile.frames[self.name].scale
	self:SetScale(scale)
end

function BarProto:UpdateHeight()
	local height = Carousel.db.profile.frames[self.name].height
	if not Carousel.db.profile.frames[self.name].locked then
		self:SetHeight(height + 16)
	else
		self:SetHeight(height)
	end
	for i, dock in pairs(self.docks) do
		dock:UpdateHeight(height)
	end
	self:UpdateWidth()
end

function BarProto:UpdateLayout()
	local db = Carousel.db.profile.frames[self.name]
	for i, dock in pairs(self.docks) do
		dock:UpdateLayout(db.hideArrows, db.blockPos, db.arrowSpace)
	end
end

function BarProto:UpdateTextIconPosition()
	local db = Carousel.db.profile.frames[self.name]
	for i, dock in pairs(self.docks) do
		dock:UpdateTextIconPosition(db.iconLeft)
	end
end

do
	local backdrop = {
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		edgeSize = 16,
		insets = {left = 5, right = 5, top = 5, bottom = 5},
		tile = true, tileSize = 16,
	}
	function BarProto:UpdateBackdrop()
		local db = Carousel.db.profile.frames[self.name]
		backdrop.edgeSize = db.borderSize
		backdrop.insets.left = db.borderInsets
		backdrop.insets.right = db.borderInsets
		backdrop.insets.top = db.borderInsets
		backdrop.insets.bottom = db.borderInsets
		self:SetBackdrop(backdrop)
		self:UpdateFrameColors()
	end
end

function BarProto:UpdateFrameColors()
	local db = Carousel.db.profile.frames[self.name]
	if not Carousel.db.profile.frames[self.name].locked then
		self:SetBackdropColor(1, 0, 0, 1)
	else
		self:SetBackdropColor(unpack(db.backdropColor))
	end
	self:SetBackdropBorderColor(unpack(db.backdropBorderColor))
end

function BarProto:UpdateFont()
	Carousel:VarifyFont(self.name)
	local db = Carousel.db.profile.frames[self.name]
	for i, dock in pairs(self.docks) do
		dock:UpdateFont(db.font, db.fontsize, db.fontoutline)
		dock:UpdateText(dock.active)
	end
end

function BarProto:UpdateTextColor()
	local db = Carousel.db.profile.frames[self.name]
	local t, l = db.textColor, db.labelColor
	self.textColor = string.format("ff%02x%02x%02x", t[1]*255, t[2]*255, t[3]*255)
	self.labelColor = string.format("ff%02x%02x%02x", l[1]*255, l[2]*255, l[3]*255)
	for i, dock in pairs(self.docks) do
		dock:UpdateTextColor(self.textColor, self.labelColor)
	end
end

function BarProto:UpdateTextAlignment()
	local db = Carousel.db.profile.frames[self.name]
	for i, dock in pairs(self.docks) do
		dock:UpdateTextAlignment(db.textAlignment)
	end
end

function BarProto:UpdateWidth()
	local db = Carousel.db.profile.frames[self.name]
	local space, maxwidth, offset = db.space, db.width, db.widthOffset
	local num = #self.docks
	if maxwidth == 0 then
		local width = 0
		if num > 0 then
			for i = 1, num do
				width = width + self.docks[i]:GetWidth() + space
			end
			width = max(width + offset - space, 32)
		else
			width = 32
		end
		if not Carousel.db.profile.frames[self.name].locked then
			width = width + 16
		end
		self:SetWidth(width)
	else
		self:SetWidth(maxwidth)
		local total = maxwidth - offset-- - space * (num-1)
		if not Carousel.db.profile.frames[self.name].locked then
			total = total - 16
		end
		local width, nextwidth
		for i,dock in ipairs(self.docks) do
			if total > dock.idealWidth then
				total = total - dock.idealWidth
				if self.docks[i+1] then
					nextwidth = self.docks[i+1].idealWidth
				else
					nextwidth = total + space + 1
				end
				if total >= nextwidth + space then
					width = dock.idealWidth
				else
					width = dock.idealWidth + total
					total = 0
				end
			elseif total > dock.arrowWidth then
				width = total
				total = 0
			else -- total >= 0 and <= dock.arrowWidth
				width = 0
			end
			if width == 0 then
				dock:Hide()
			else
				dock:Show()
				dock:SetWidth(width)
				dock.display:SetWidth(width - dock.arrowWidth)
			end
			total = total - space
		end
	end
end

function BarProto:UpdateFixedWidth()
	if Carousel.db.profile.frames[self.name].width == 0 then
		for i, dock in pairs(self.docks) do
			dock:Show()
			dock:ResizeToText()
		end
	end
	self:UpdateWidth()
end

function BarProto:UpdateOrder()
	local order = Carousel.db.profile.frames[self.name].objectOrder
	for dock, tab in pairs(order) do
		self.docks[dock]:UpdateOrder(tab)
	end
end

function BarProto:UpdateSpace()
	self:AnchorDocks()
	self:UpdateWidth()
end

function BarProto:UpdateLock()
	local lock = Carousel.db.profile.frames[self.name].locked
	if lock then
		self:Lock()
	else
		self:Unlock()
	end
end

function BarProto:Unlock()
	self:SetMovable(true)
	self:RegisterForDrag("LeftButton")
	self:UpdateHeight()
	--self:UpdateWidth()
	self:UpdateFrameColors()
	self:AnchorDocks()
end

function BarProto:Lock()
	self:SetMovable(false)
	self:RegisterForDrag(nil)
	self:SetScript("OnUpdate", nil)
	self:UpdateHeight() -- Calls update width
	--self:UpdateWidth()
	self:UpdateFrameColors()
	self:UpdateAnchor()
	self:AnchorDocks()
end

function BarProto.OnDragStart(self)
	self:StartMoving()
end

do
	local function OnUpdate(self)
		self:SetScript("OnUpdate", nil)
		self:UpdateDragPosition()
	end
	function BarProto.OnDragStop(self)
		self:StopMovingOrSizing()
		self:SetScript("OnUpdate", OnUpdate)
	end
end

function BarProto:UpdateDragPosition()
	local point, _, anchor, x, y = self:GetPoint(1)
	if not point then return end
	--local s = self:GetEffectiveScale()
	local fs = self:GetScale()
	local db = Carousel.db.profile.frames[self.name]
	local anchorSide = db.anchorSide
	if anchorSide == "RIGHT" and (point ~= "RIGHT" or anchor ~= "RIGHT") then
		x = self:GetRight() - GetScreenWidth()/fs
		y = select(2, self:GetCenter()) - (GetScreenHeight()/2)/fs
	elseif anchorSide == "LEFT" and (point ~= "LEFT" or anchor ~= "LEFT") then
		x = self:GetLeft()
		y = select(2, self:GetCenter()) - (GetScreenHeight()/2)/fs
	elseif anchorSide == "CENTER" and (point ~= "CENTER" or anchor ~= "CENTER") then
		x = select(1, self:GetCenter()) - (GetScreenWidth()/2)/fs
		y = select(2, self:GetCenter()) - (GetScreenHeight()/2)/fs
	end
	db.frameX = x * db.scale
	db.frameY = y * db.scale
end
